package org.artifactory.federation;

import org.artifactory.descriptor.authentication.AuthenticationToken;
import org.artifactory.descriptor.repo.FederatedMember;
import org.artifactory.descriptor.repo.FederatedRepoDescriptor;
import org.artifactory.repo.RepoDetails;
import org.artifactory.rest.exception.ForbiddenException;
import org.artifactory.spring.ContextCreationListener;
import org.artifactory.spring.EntitlementsEventListener;

import java.util.Collection;
import java.util.List;
import java.util.Set;

/**
 * @author Alexei Vainshtein
 */
public interface FederatedRepoService extends EntitlementsEventListener, ContextCreationListener {

    /**
     * Create remote federated members.
     */
    void createRecreateOrRemoveRemoteMembers(FederatedRepoDescriptor newDescriptor,  Set<FederatedMember> oldMembers);

    /**
     * Fetch from each remote url the existed federated repos.
     *
     * @param urls - The remote urls to check
     * @return a list that contains a json with url and federation repos of that url.
     */
    List<FederationReposResponse> getReposFromRemote(List<String> urls, String proxyName);

    /**
     * @return a list of all federated repositories and the info per repo
     */
    List<RepoDetails> getRepos();

    /**
     * @return the token that associated with remote url and the key provided
     */
    String getToken(String remoteUrl, String key);

    /**
     * Validates the given repository
     *
     * @param key                      - repo name
     * @param clientBaseUrl            - base url of the requesting server
     * @param timeFromRequestingServer - the current time of the server that is requesting the validation.
     */
    void validate(String key, String clientBaseUrl, long timeFromRequestingServer);

    /**
     * Checks if the license provided and subscription (if AOL) are enterprise and above.
     * {@link ForbiddenException} is thrown if the license/subscription are not enterprise and above.
     */
    void validateFederationLicenseInstalled();

    /**
     * Validates remote url
     *
     * @param fullUrl - The remote url including the repository to validate
     */
    void validateRemoteServer(String fullUrl, String proxyName);

    /**
     * Migrates local repository to a federated repository
     *
     * @param repoKey - The repository name that need to be migrated.
     */
    default void migrate(String repoKey) {

    }

    default void triggerFullSync(String repoKey, String mirrorUrl) {

    }

    default void configSync(String repoKey) {

    }

    /**
     * Returns whether federated repo is ready to use
     */
    default boolean isActive(String repoKey) {
        return false;
    }

    void addInconsistentRepoMember(String memberUrl, String repoKey);

    Set<String> getInconsistentRepos();

    void removeInconsistentRepoMember(String memberUrl);

    void removeRepoFromRemoteMembers(FederatedRepoDescriptor federatedDescriptor);

    void saveAuthenticationMasterToken(String tokenId, String tokenType, String scope,
            String baseUrl, String exchangeUrl, String accessToken, Long expiresIn);

    Collection<AuthenticationToken> getPairedAuthenticationTokens();
}
