package org.artifactory.queue.services;

import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;
import lombok.extern.slf4j.Slf4j;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.HaAddon;
import org.artifactory.common.ConstantValues;
import org.artifactory.queue.QueueStatus;
import org.artifactory.storage.db.event.entity.NodeEventCursorType;
import org.artifactory.storage.db.event.model.NodeEventCursor;
import org.artifactory.storage.db.event.service.NodeEventCursorService;
import org.artifactory.storage.event.EventsService;
import org.jetbrains.annotations.NotNull;
import org.jfrog.common.concurrency.ConflictsGuard;
import org.springframework.stereotype.Service;

import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;

@Slf4j
@Service
public class PersistentQueueCursorServiceImpl implements PersistentQueueCursorService {

    public static final String CREATE_CURSOR_CONFLICTS_GUARD = "create-cursor";
    private final NodeEventCursorService nodeEventCursorService;
    private final LoadingCache<String, Long> repoMinEventIdCache;
    private final ConflictsGuard<String> createCursorConflictsGuard;
    private final EventsService eventsService;

    public PersistentQueueCursorServiceImpl(NodeEventCursorService nodeEventCursorService,
            EventsService eventsService, AddonsManager addonsManager) {
        this.nodeEventCursorService = nodeEventCursorService;
        this.eventsService = eventsService;
        this.repoMinEventIdCache = CacheBuilder.newBuilder()
                .expireAfterAccess(ConstantValues.persistentQueueRepoMinEventCacheExpireSec.getInt(),
                        TimeUnit.SECONDS)
                .concurrencyLevel(Runtime.getRuntime().availableProcessors())
                .build(new CacheLoader<>() {
                    @Override
                    public Long load(@NotNull String repoKey) {
                        return eventsService.getFirstEventLogId(repoKey);
                    }
                });

        this.createCursorConflictsGuard =
                addonsManager.addonByType(HaAddon.class).getConflictsGuard(CREATE_CURSOR_CONFLICTS_GUARD);
    }

    /**
     * Create marker if not exists. Protected by db lock.
     *
     * @param repoKey    the repo key of the marker
     * @param cursorName rhe cursor name
     * @param type       the cursor type
     */
    @Override
    public void createMarkerIfNotExists(String repoKey, String cursorName, NodeEventCursorType type) {
        if (nodeEventCursorService.cursorForOperator(cursorName) == null) {
            try {
                if (createCursorConflictsGuard.tryToLock(cursorName, 10L, TimeUnit.SECONDS)) {
                    try {
                        if (nodeEventCursorService.cursorForOperator(cursorName) == null) {
                            createCursor(cursorName, type);
                        }
                    } finally {
                        createCursorConflictsGuard.unlock(cursorName);
                    }
                }
            } catch (InterruptedException e) {
                Thread.currentThread().interrupt();
                throw new RuntimeException("Failed to create cursor", e);
            }
        }
    }

    @Override
    public void updateMarker(String cursorName, long newMarker) {
        nodeEventCursorService.updateCursor(cursorName, newMarker);
    }

    @Override
    public void resetMarker(String repoKey, String cursorName) {
        long lastEventLogId = eventsService.getLastEventLogId();
        log.info("reset the {} to {}", cursorName, lastEventLogId);
        updateMarker(cursorName, lastEventLogId);
    }

    @Override
    public void deleteMarker(String cursorName) {
        log.info("delete the cursor {}", cursorName);
        nodeEventCursorService.deleteCursor(cursorName);
    }

    @Override
    public long getCurrentMarker(String repoKey, String cursorName, NodeEventCursorType type) {
        NodeEventCursor cursor = nodeEventCursorService.cursorForOperator(cursorName);
        if (cursor == null) {
            //create cursor for 1st time
            return createCursor(cursorName, type);
        }
        return cursor.getEventMarker();
    }

    private long createCursor(String cursorName, NodeEventCursorType type) {
        nodeEventCursorService.insertCursor(cursorName, -1, type);
        log.info("Cursor {} created with marker -1", cursorName);
        return -1;
    }

    @Override
    public QueueStatus getStatus(String repoKey, String cursorName) {
        //this is the logic returned Status
        //no cursor exist                => N/A
        //marker=X && firstEvent<X => OK
        //marker=X && firstEvent=X => OK
        //marker=X && firstEvent=X+1 => OK
        //marker=X && firstEvent=X+2 => DataLost
        NodeEventCursor cursor = nodeEventCursorService.cursorForOperator(cursorName);
        if (cursor == null) {
            return QueueStatus.NOT_AVAILABLE;
        }
        long marker = cursor.getEventMarker();
        long minEventId = getRepoMinEventId(repoKey);

        if (marker == -1) {
            log.info("Found a new marker for {}", cursorName);
            return QueueStatus.NEW_MARKER;
        }

        if (marker + 1 < minEventId) {
            log.warn("Data Lost: {}, marker: {}, minEventId:{}", cursorName, marker, minEventId);
            return QueueStatus.DATA_LOST;
        } else {
            return QueueStatus.OK;
        }
    }

    private long getRepoMinEventId(@NotNull String repoKey) {
        try {
            return repoMinEventIdCache.get(repoKey);
        } catch (ExecutionException e) {
            throw new RuntimeException("Failed to get repo min eventId", e);
        }
    }
}
