/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.work.queue;

import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.HaAddon;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.repo.AsyncWorkQueueCustomExecutor;
import org.artifactory.api.repo.AsyncWorkQueueProviderService;
import org.artifactory.api.repo.WorkItem;
import org.artifactory.api.repo.WorkQueue;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.common.home.TestArtifactoryHome;
import org.artifactory.concurrent.NonStandardLockCategory;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.repo.FederatedRepoDescriptor;
import org.artifactory.mbean.MBeanRegistrationService;
import org.artifactory.spring.InternalArtifactoryContext;
import org.artifactory.work.queue.mbean.WorkQueueMBean;
import org.easymock.EasyMock;
import org.mockito.Mockito;
import org.springframework.scheduling.concurrent.ConcurrentTaskExecutor;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.Method;
import java.nio.file.Files;
import java.util.HashMap;
import java.util.Map;

import static org.easymock.EasyMock.anyObject;
import static org.easymock.EasyMock.createNiceMock;
import static org.easymock.EasyMock.eq;
import static org.easymock.EasyMock.expect;
import static org.easymock.EasyMock.expectLastCall;
import static org.easymock.EasyMock.replay;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertNotEquals;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertNull;
import static org.testng.Assert.assertTrue;

/**
 * @author Shay Bagants
 */
@Test
public class AsyncWorkQueueServiceTest {
    private File home;
    private HaAddon haAddon;

    @BeforeMethod
    private void setup() throws IOException {
        home = Files.createTempDirectory(this.getClass().getSimpleName()).toFile();
        ArtifactoryHome artifactoryHome = new TestArtifactoryHome(home);
        artifactoryHome.initArtifactorySystemProperties();
        ArtifactoryHome.bind(artifactoryHome);

        haAddon = createNiceMock(HaAddon.class);
        expect(haAddon.getConflictsGuard(anyObject(String.class))).andReturn(null).anyTimes();

        AddonsManager addonsManager = createNiceMock(AddonsManager.class);
        expect(addonsManager.addonByType(HaAddon.class)).andReturn(haAddon).anyTimes();

        MBeanRegistrationService mbeanService = createNiceMock(MBeanRegistrationService.class);
        mbeanService.register(anyObject(WorkQueueMBean.class), anyObject(String.class), anyObject(String.class));
        expectLastCall().anyTimes();

        InternalArtifactoryContext context = EasyMock.createMock(InternalArtifactoryContext.class);
        expect(context.beanForType(AddonsManager.class)).andReturn(addonsManager).anyTimes();
        expect(context.beanForType(MBeanRegistrationService.class)).andReturn(mbeanService).times(3);

        EasyMock.replay(context, addonsManager, haAddon, mbeanService);
        ArtifactoryContextThreadBinder.bind(context);
    }

    @AfterMethod
    private void cleanup() {
        ArtifactoryContextThreadBinder.unbind();
        home.delete();
    }

    // Request the queue using two different method names and expecting both to return the same exact queue object
    public void testGetQueue() throws NoSuchMethodException {
        EasyMock.reset(haAddon);
        expect(haAddon.getConflictsGuard(eq(NonStandardLockCategory.FED_REPO_FULL_SYNC.getName()))).andReturn(null)
                .times(2);
        replay(haAddon);
        WorkQueueTestMethodsService workQueueTestMethodsService = new WorkQueueTestMethodsService();
        Method asyncMethod = workQueueTestMethodsService.getClass().getMethod("calculateDebianMetadataInternalAsync");
        Method syncMethod = workQueueTestMethodsService.getClass().getMethod("calculateDebianMetadataInternalSync");
        Method mvnMethod = workQueueTestMethodsService.getClass().getMethod("calculateMavenMetadataAsync");
        Method fullSyncMirrorMethod = workQueueTestMethodsService.getClass().getMethod("runFullSyncAsync");
        Method triggerFullSync = workQueueTestMethodsService.getClass().getMethod("triggerFullSync");

        AsyncWorkQueueServiceImpl workQueueService = new AsyncWorkQueueServiceImpl();
        // request workqueue for both methods, expecting both to return the same queue object
        WorkQueue<WorkItem> asyncQueue = workQueueService.getWorkQueue(asyncMethod, workQueueTestMethodsService);
        WorkQueue<WorkItem> syncQueue = workQueueService.getWorkQueue(syncMethod, workQueueTestMethodsService);
        WorkQueue<WorkItem> mvnQueue = workQueueService.getWorkQueue(mvnMethod, workQueueTestMethodsService);
        WorkQueue<WorkItem> fullSyncMirrorQueue =
                workQueueService.getWorkQueue(fullSyncMirrorMethod, workQueueTestMethodsService);
        WorkQueue<WorkItem> triggerFullSyncWorkQueue =
                workQueueService.getWorkQueue(triggerFullSync, workQueueTestMethodsService);
        // the comparision should not be by the content, but by the reference!
        assertTrue(asyncQueue == syncQueue);
        // make sure the maven queue is not the same as the debian
        assertTrue(asyncQueue != mvnQueue);

        // validate the queues map content
        Map<String, WorkQueue<WorkItem>> queues = workQueueService.getExistingWorkQueues();
        assertEquals(queues.size(), 5);
        assertNotEquals(queues.get("calculateDebianMetadataInternalAsync"), null);
        // the comparision should not be by the content, but by the reference! we expect the queue to have two keys,
        //each key points on the same object
        assertTrue(queues.get("calculateDebianMetadataInternalAsync") ==
                queues.get("calculateDebianMetadataInternalSync"));
        // make sure the maven queue is not the same as the debian
        assertTrue(queues.get("mvnQueue") != queues.get("calculateDebianMetadataInternalAsync"));
        assertEquals(fullSyncMirrorQueue.getMaxNumberOfWorkers(), ConstantValues.mirrorFullSyncWorkers.getInt());
        assertEquals(triggerFullSyncWorkQueue.getMaxNumberOfWorkers(), ConstantValues.mirrorFullSyncWorkers.getInt());
        assertEquals(fullSyncMirrorQueue.getName(), "Full sync triggering");
        assertEquals(triggerFullSyncWorkQueue.getName(), "Full sync triggering");
        EasyMock.verify(haAddon);
    }

    public void testCustomExecutor() throws NoSuchMethodException {
        WorkQueueTestMethodsService workQueueTestMethodsService = new WorkQueueTestMethodsService();
        Method asyncMethod = workQueueTestMethodsService.getClass().getMethod("calculateDebianMetadataInternalAsync");

        AsyncWorkQueueServiceImpl workQueueService = new AsyncWorkQueueServiceImpl();

        CentralConfigService centralConfigService = mock(CentralConfigService.class);
        CentralConfigDescriptor descriptor = mock(CentralConfigDescriptor.class);
        Mockito.when(centralConfigService.getDescriptor()).thenReturn(descriptor);

        FederatedRepoDescriptor fed1 = mock(FederatedRepoDescriptor.class);
        when(fed1.getMaxNumberOfThreads()).thenReturn(0);
        FederatedRepoDescriptor fed2 = mock(FederatedRepoDescriptor.class);
        when(fed2.getMaxNumberOfThreads()).thenReturn(20);

        Map<String, FederatedRepoDescriptor> feds = new HashMap<>();
        feds.put("repo1", fed1);
        feds.put("repo2", fed2);
        when(descriptor.getFederatedRepositoriesMap()).thenReturn(feds);

        WorkQueue<WorkItem> asyncQueue1;
        WorkQueue<WorkItem> asyncQueue2;
        ConcurrentTaskExecutor executor = mock(ConcurrentTaskExecutor.class);
        try {
            AsyncWorkQueueProviderService.CustomExecutor.set(AsyncWorkQueueCustomExecutor.builder()
                    .nameSuffix("test")
                    .repoKey("repo1")
                    .executor(executor)
                    .threadsCount(() -> 4)
                    .build());
            asyncQueue1 = workQueueService.getWorkQueue(asyncMethod, workQueueTestMethodsService);
            asyncQueue2 = workQueueService.getWorkQueue(asyncMethod, workQueueTestMethodsService);
        } finally {
            AsyncWorkQueueProviderService.CustomExecutor.remove();
        }
        assertTrue(asyncQueue1.getName().contains("test"));
        assertTrue(asyncQueue1.getName().contains("repo1"));
        assertNotNull(asyncQueue1.customExecutor());
        assertEquals(asyncQueue1.customExecutor().getExecutor(), executor);
        assertEquals(asyncQueue1.availablePermits(), 4);
        assertEquals(asyncQueue2, asyncQueue1);

        WorkQueue<WorkItem> asyncQueue3;
        try {
            AsyncWorkQueueProviderService.CustomExecutor.set(AsyncWorkQueueCustomExecutor.builder()
                    .nameSuffix("test")
                    .repoKey("repo2")
                    .executor(executor)
                    .threadsCount(() -> fed2.getMaxNumberOfThreads())
                    .build());
            asyncQueue3 = workQueueService.getWorkQueue(asyncMethod, workQueueTestMethodsService);
        } finally {
            AsyncWorkQueueProviderService.CustomExecutor.remove();
        }

        assertTrue(asyncQueue3.getName().contains("test"));
        assertTrue(asyncQueue3.getName().contains("repo2"));
        assertNotNull(asyncQueue3.customExecutor());
        assertEquals(asyncQueue3.customExecutor().getExecutor(), executor);
        assertEquals(asyncQueue3.availablePermits(), 20);
        assertNotEquals(asyncQueue3, asyncQueue1);

        WorkQueue<WorkItem> asyncQueue4 = workQueueService.getWorkQueue(asyncMethod, workQueueTestMethodsService);
        assertFalse(asyncQueue4.getName().contains("test"));
        assertNull(asyncQueue4.customExecutor());
        assertEquals(asyncQueue4.availablePermits(), ConstantValues.debianMetadataCalculationWorkers.getInt());
        assertNotEquals(asyncQueue4, asyncQueue1);
    }

    // we only use this class with these two methods to pass it into the AsyncWorkQueueServiceImpl when requesting a
    //workQueue
    public static class WorkQueueTestMethodsService {

        public void calculateDebianMetadataInternalAsync() {
        }

        public void calculateDebianMetadataInternalSync() {
        }

        public void calculateMavenMetadataAsync() {
        }

        public void runFullSyncAsync() {
        }
        public void triggerFullSync() {
        }
    }
}
