#!/bin/bash
set -e
OPTIND=1

display_usage(){
    echo "THIS SCRIPT IS USED TO PUSH FILES TO REMOTE GITHUB REPOSITORY"
    echo "-n Path of the file - mandatory (Absolute path in which the file is located)"
    echo "-r Specify repository - mandatory (repository name without the project name)."
    echo "-b Specify git branch (make sure the branch exists in remote) - optional (otherwise, will push to master)"
    echo "-d Specify the directory inside the repository - optional (otherwise, will push to the main directory)"
    echo "-p Specify the path for the private ssh key file - optional (otherwise, will take '~/.ssh/id_rsa')"
    echo "EXAMPLE: ./git_pusher.sh -n /path/to/file -r some_repo -d path/to/repository/directory"
}

error() {
  echo "ERROR: $1"
  exit 1
}

validateFlags() {
    local repository_name="$1"
    shift 1
    local arr=("$@")

    if [[ -z "${repository_name}" ]]; then
        error "Repository name was not given, exiting..."
    fi

    if [[ "${#arr[@]}" -eq 0 ]]; then
        error "source files paths were not given, exiting..."
    fi
}

source_files_paths=
repository_name=
git_branch="master"
target_directory="$(pwd)"
target_repo_path=
private_key_path=~/.ssh/id_rsa

while getopts "p:r:n:d:b:h" option; do
  case "$option" in
  h)
    display_usage
    exit 0
    ;;
  n)
    source_files_paths="${OPTARG}"
    ;;
  r)
    repository_name="${OPTARG}"
    ;;
  d)
    target_repo_path="${OPTARG}"
    ;;
  b)
    git_branch="${OPTARG}"
    ;;
  p)
    private_key_path="${OPTARG}"
    ;;
  *)
    display_usage
    exit 1
    ;;
  esac
done

ssh_command="ssh -i ${private_key_path}"
IFS=',' read -r -a source_files_paths <<< "$source_files_paths"
validateFlags "${repository_name}" "${source_files_paths[@]}"
# project name will always be jfrog
project_name="jfrog"
# builds the remote url for git commands
url="git@github.com:${project_name}/${repository_name}.git"
# creates/update a repository and checkout to the relevant branch
if [[ -d "${repository_name}" ]]; then
    cd "${repository_name}"
    git config core.sshCommand "${ssh_command}"
    git checkout "${git_branch}"
    git pull
else
    git -c core.sshCommand="${ssh_command}" clone "${url}"
    cd "${repository_name}"
    git checkout "${git_branch}"
    git config core.sshCommand "${ssh_command}"
fi
# adds '/' to target_repo_path if needed
[[ "${target_repo_path: -1}" == "/" || -z "${target_repo_path}" ]] || target_repo_path="${target_repo_path}/"
# builds the absolute path of the target directory
target_directory="${target_directory}/${repository_name}/${target_repo_path}"
# creates a directory if it does not exists
mkdir -p "${target_directory}"
# git config in order to make git actions
git config user.name "jfrog"
git config user.email "deployer@jfrog.com"
# loop through all the files
for source_file_path in "${source_files_paths[@]}"
do
    # gets the file name to commit
    file_name=$(basename "${source_file_path}")
    # copy the source file to the target directory path
    cp -f "${source_file_path}" "${target_directory}"
    # adds the file to git
    git add "${target_repo_path}${file_name}"
done
# commits the file(s)
files_to_commit=$(git diff --name-only --cached)
if [[ -n "${files_to_commit}" ]]; then
  git commit -m "Added/Updated the following files: ${files_to_commit}"
  # pushes the file(s) to remote
  git push origin "${git_branch}"
  # information regarding the push
  echo "Github push ended successfully!"
  echo "Repository: ${project_name}/${repository_name}"
  echo "Branch: ${git_branch}"
  echo "Files: ${files_to_commit}"
else
  echo "Nothing to commit, exiting..."
fi
