package org.artifactory.ui.rest.service.admin.configuration.repositories.util.builder;

import com.google.common.collect.Maps;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.webstart.ArtifactWebstartAddon;
import org.artifactory.aol.AolService;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.common.property.ArtifactorySystemProperties;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.descriptor.repo.CdnRedirectRepoConfig;
import org.artifactory.descriptor.repo.FederatedRepoDescriptor;
import org.artifactory.descriptor.repo.HttpRepoDescriptor;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.ProxyDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.spring.InternalArtifactoryContext;
import org.artifactory.ui.rest.model.admin.configuration.repository.federation.FederatedRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.local.LocalRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.remote.RemoteCacheRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.remote.RemoteRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.typespecific.DockerTypeSpecificConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.typespecific.GoTypeSpecificConfigModel;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.util.List;
import java.util.Map;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertNull;
import static org.testng.Assert.assertTrue;

/**
 * Author: Ivan Perepelytsya
 * Date: 6/2/20.
 */
@Test
public class RepoConfigModelBuilderTest {

    @BeforeMethod
    public void bindArtifactoryHome() {
        InternalArtifactoryContext context = mock(InternalArtifactoryContext.class);
        ArtifactoryContextThreadBinder.bind(context);
        AolService aolService = mock(AolService.class);
        when(aolService.isAol()).thenReturn(false);
        when(context.beanForType(AolService.class)).thenReturn(aolService);
        final ArtifactorySystemProperties artifactorySystemProperties = mock(ArtifactorySystemProperties.class);
        final ArtifactoryHome artifactoryHome = mock(ArtifactoryHome.class);
        when(artifactoryHome.getArtifactoryProperties()).thenReturn(artifactorySystemProperties);
        when(artifactorySystemProperties.getProperty(any())).thenReturn("");
        ArtifactoryHome.bind(artifactoryHome);
    }

    @AfterMethod
    public void unbindArtifactoryHome() {
        ArtifactoryHome.unbind();
    }

    @Test
    public void testPopulateLocalDescriptorCdn() {
        final RepoConfigModelBuilder repoConfigModelBuilder = new RepoConfigModelBuilder();
        repoConfigModelBuilder.setCentralConfig(getCentralConfigService());

        final LocalRepoDescriptor localRepoDescriptor = new LocalRepoDescriptor();
        localRepoDescriptor.setCdnRedirectConfig(getEnabledCdnConfig());

        final LocalRepositoryConfigModel localRepositoryConfigModel =
                new LocalRepositoryConfigModel();
        repoConfigModelBuilder.populateLocalDescriptorValuesToModel(localRepoDescriptor,
                localRepositoryConfigModel);

        assertTrue(localRepositoryConfigModel.getAdvanced().getCdnRedirectRepoConfig().isEnabled());
    }

    @Test
    public void testPopulateRemoteDescriptorCdn() {
        final RepoConfigModelBuilder repoConfigModelBuilder = getRepoConfigModelBuilder();

        final HttpRepoDescriptor remoteRepoDescriptor = new HttpRepoDescriptor();
        remoteRepoDescriptor.setCdnRedirectConfig(getEnabledCdnConfig());

        final RemoteRepositoryConfigModel remoteRepositoryConfigModel = new RemoteRepositoryConfigModel();
        repoConfigModelBuilder
                .populateRemoteRepositoryConfigValuesToModel(remoteRepoDescriptor, remoteRepositoryConfigModel);

        assertTrue(remoteRepositoryConfigModel.getAdvanced().getCdnRedirectRepoConfig().isEnabled());
    }

    @Test
    public void testPopulateLocalDescriptorNullCdn() {

        final RepoConfigModelBuilder repoConfigModelBuilder = new RepoConfigModelBuilder();
        repoConfigModelBuilder.setCentralConfig(getCentralConfigService());

        final LocalRepoDescriptor localRepoDescriptor = new LocalRepoDescriptor();
        localRepoDescriptor.setCdnRedirectConfig(null);

        final LocalRepositoryConfigModel localRepositoryConfigModel =
                new LocalRepositoryConfigModel();
        repoConfigModelBuilder.populateLocalDescriptorValuesToModel(localRepoDescriptor,
                localRepositoryConfigModel);
        assertNull(localRepositoryConfigModel.getAdvanced().getCdnRedirectRepoConfig());
    }


    @Test
    public void testPopulateRemoteDescriptorNullCdn() {
        final RepoConfigModelBuilder repoConfigModelBuilder = getRepoConfigModelBuilder();

        final HttpRepoDescriptor remoteRepoDescriptor = new HttpRepoDescriptor();
        remoteRepoDescriptor.setCdnRedirectConfig(null);

        final RemoteRepositoryConfigModel remoteRepositoryConfigModel = new RemoteRepositoryConfigModel();
        repoConfigModelBuilder
                .populateRemoteRepositoryConfigValuesToModel(remoteRepoDescriptor, remoteRepositoryConfigModel);

        assertNull(remoteRepositoryConfigModel.getAdvanced().getCdnRedirectRepoConfig());
    }

    @Test(description = "Tests repository stored RetrievalCachePeriod, UI gets this for the Advanced -> Cache section" +
            "and displays this on RetrievalCachePeriod field")
    public void testPopulateRemoteRepositoryDockerConfigValuesToModel() {
        RepoConfigModelBuilder repoConfigModelBuilder = new RepoConfigModelBuilder();
        HttpRepoDescriptor httpRepoDescriptor = new HttpRepoDescriptor();
        httpRepoDescriptor.setRetrievalCachePeriodSecs(500L);

        DockerTypeSpecificConfigModel remoteTypeSpecific = (DockerTypeSpecificConfigModel) repoConfigModelBuilder
                .createRemoteTypeSpecific(RepoType.Docker, httpRepoDescriptor);

        assertEquals((long) remoteTypeSpecific.getRetrievalCachePeriodSecs(), 500L);
    }

    public void testPopulateRemoteRepositoryGoConfigValuesToModel() {
        RepoConfigModelBuilder repoConfigModelBuilder = new RepoConfigModelBuilder();
        HttpRepoDescriptor httpRepoDescriptor = new HttpRepoDescriptor();
        httpRepoDescriptor.setMissedRetrievalCachePeriodSecs(300L);

        GoTypeSpecificConfigModel remoteTypeSpecific = (GoTypeSpecificConfigModel) repoConfigModelBuilder
                .createRemoteTypeSpecific(RepoType.Go, httpRepoDescriptor);

        assertEquals((long) remoteTypeSpecific.getMissedRetrievalCachePeriodSecs(), 300L);
    }

    @Test
    public void testResolveLocalP2RepoFromUrl() {
        RepoConfigModelBuilder repoConfigModelBuilder = new RepoConfigModelBuilder();
        CentralConfigService centralConfigService = getCentralConfigService();
        String repoName = "repo";
        createMutableDescriptorMock(centralConfigService, repoName);
        repoConfigModelBuilder.setCentralConfig(centralConfigService);

        assertNotNull(repoConfigModelBuilder.resolveLocalP2RepoFromUrl("local://" + repoName));
    }

    @Test
    public void testPopulateFederatedDescriptorValuesToModel() {
        RepoConfigModelBuilder repoConfigModelBuilder = new RepoConfigModelBuilder();
        repoConfigModelBuilder.setCentralConfig(getCentralConfigService());

        FederatedRepoDescriptor repoDescriptor = new FederatedRepoDescriptor();
        ProxyDescriptor proxyDescriptor = new ProxyDescriptor();
        proxyDescriptor.setKey("proxy");
        repoDescriptor.setProxy(proxyDescriptor);
        FederatedRepositoryConfigModel model = new FederatedRepositoryConfigModel();
        repoConfigModelBuilder.populateFederatedDescriptorValuesToModel(repoDescriptor, model);
        assertNotNull(model.getAdvanced().getProxy());
        assertEquals(model.getAdvanced().getProxy(), proxyDescriptor.getKey());
    }

    @Test
    public void testCreateCacheConfig() {
        RepoConfigModelBuilder repoConfigModelBuilder = new RepoConfigModelBuilder();
        HttpRepoDescriptor repoDescriptor = new HttpRepoDescriptor();
        RemoteCacheRepositoryConfigModel cacheConfig = repoConfigModelBuilder.createCacheConfig(repoDescriptor);
        long metadataRetrievalTimeoutSecsCache = cacheConfig.getMetadataRetrievalTimeoutSecs();
        long metadataRetrievalTimeoutSecsDescriptor = repoDescriptor.getMetadataRetrievalTimeoutSecs();
        assertEquals(metadataRetrievalTimeoutSecsCache, metadataRetrievalTimeoutSecsDescriptor);

        repoDescriptor.setMetadataRetrievalTimeoutSecs(20);
        cacheConfig = repoConfigModelBuilder.createCacheConfig(repoDescriptor);
        metadataRetrievalTimeoutSecsCache = cacheConfig.getMetadataRetrievalTimeoutSecs();
        metadataRetrievalTimeoutSecsDescriptor = repoDescriptor.getMetadataRetrievalTimeoutSecs();
        assertEquals(metadataRetrievalTimeoutSecsCache, metadataRetrievalTimeoutSecsDescriptor);
    }

    private void createMutableDescriptorMock(CentralConfigService centralConfigService, String repoName) {
        MutableCentralConfigDescriptor descriptor = mock(MutableCentralConfigDescriptor.class);
        Map<String, LocalRepoDescriptor> reposMap = Maps.newLinkedHashMap();
        FederatedRepoDescriptor federatedRepo = new FederatedRepoDescriptor();
        federatedRepo.setKey(repoName);
        reposMap.put(repoName, federatedRepo);
        when(descriptor.getLocalAndFederatedRepositoriesMap()).thenReturn(reposMap);
        when(centralConfigService.getMutableDescriptor()).thenReturn(descriptor);
    }

    private CentralConfigService getCentralConfigService() {
        final CentralConfigService centralConfigService = mock(CentralConfigService.class);
        final CentralConfigDescriptor centralConfigDescriptor = mock(CentralConfigDescriptor.class);

        when(centralConfigService.getDescriptor()).thenReturn(centralConfigDescriptor);
        return centralConfigService;
    }

    private CdnRedirectRepoConfig getEnabledCdnConfig() {
        CdnRedirectRepoConfig cdnRedirectRepoConfig = new CdnRedirectRepoConfig();
        cdnRedirectRepoConfig.setEnabled(true);
        return cdnRedirectRepoConfig;
    }


    private RepoConfigModelBuilder getRepoConfigModelBuilder() {
        final RepoConfigModelBuilder repoConfigModelBuilder = new RepoConfigModelBuilder();
        repoConfigModelBuilder.setCentralConfig(getCentralConfigService());
        final AddonsManager addonsManager = mock(AddonsManager.class);
        final ArtifactWebstartAddon artifactWebstartAddon = mock(ArtifactWebstartAddon.class);
        when(artifactWebstartAddon.getSslCertNames()).thenReturn(List.of());
        when(addonsManager.addonByType(ArtifactWebstartAddon.class)).thenReturn(artifactWebstartAddon);
        repoConfigModelBuilder.setAddonsManager(addonsManager);
        return repoConfigModelBuilder;
    }
}