package org.artifactory.ui.rest.service.artifacts.browse.generic;

import org.apache.http.HttpStatus;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.CoreAddons;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.repo.BaseBrowsableItem;
import org.artifactory.api.repo.BrowsableItem;
import org.artifactory.api.repo.BrowsableItemCriteria;
import org.artifactory.api.repo.RepositoryBrowsingService;
import org.artifactory.api.repo.RootNodesFilterResult;
import org.artifactory.api.repo.exception.FolderExpectedException;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.VirtualRepoDescriptor;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.ArtifactoryRestResponse;
import org.artifactory.test.ArtifactoryHomeBoundJUnitTest;
import org.artifactory.ui.rest.model.artifacts.browse.generic.NativeBrowserChild;
import org.artifactory.ui.rest.model.artifacts.browse.generic.NativeBrowserModel;
import org.artifactory.ui.rest.resource.artifacts.browse.generic.NativeBrowserResource;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;
import org.mockito.stubbing.Answer;
import org.springframework.mock.web.MockHttpServletResponse;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import static org.artifactory.common.ConstantValues.uiHideChecksums;
import static org.mockito.ArgumentMatchers.any;

public class BrowseNativeModelServiceTest extends ArtifactoryHomeBoundJUnitTest {

    @Mock
    private InternalRepositoryService repositoryService;
    @Mock
    private RepositoryBrowsingService repoBrowsingService;
    @Mock
    private AuthorizationService authorizationService;
    @Mock
    private AddonsManager addonsManager;
    @Mock
    private CoreAddons coreAddons;
    @Mock
    private ArtifactoryContext artifactoryContext;
    @InjectMocks
    private BrowseNativeModelService browseNativeService;

    @BeforeEach
    public void setUp() {
        MockitoAnnotations.openMocks(this);
        ArtifactoryContextThreadBinder.bind(artifactoryContext);
        Mockito.when(artifactoryContext.beanForType(AddonsManager.class)).thenReturn(addonsManager);
        Mockito.when(addonsManager.addonByType(CoreAddons.class)).thenReturn(coreAddons);
        Mockito.when(coreAddons.getListBrowsingVersion()).thenReturn("1.0.0");
        ArtifactoryHome.get().getArtifactoryProperties().setProperty(uiHideChecksums.getPropertyName(), "FALSE");
    }

    @AfterEach
    public void tearDown() {
        ArtifactoryHome.unbind();
    }

    @Test
    public void testWithEmptyRepoPath() {
        TestRestRequest restRequest = Mockito.mock(TestRestRequest.class);
        ArtifactoryRestResponse<NativeBrowserModel> response = new ArtifactoryRestResponse<>();
        browseNativeService.execute(restRequest, response);

        Assertions.assertEquals(HttpStatus.SC_NOT_FOUND, response.getResponseCode());
    }

    @Test
    public void testWithEmptyRepoPathAndAnonUser() {
        TestRestRequest restRequest = Mockito.mock(TestRestRequest.class);
        Mockito.when(authorizationService.isAnonAccessEnabled()).thenReturn(true);
        Mockito.when(authorizationService.isAnonymous()).thenReturn(true);
        ArtifactoryRestResponse<NativeBrowserModel> response = new ArtifactoryRestResponse<>();
        browseNativeService.execute(restRequest, response);

        Assertions.assertEquals(HttpStatus.SC_UNAUTHORIZED, response.getResponseCode());
    }

    @Test
    public void testWithUnauthorizedUserAnonAccessEnabled() {
        Mockito.when(authorizationService.canRead(any(RepoPath.class))).thenReturn(false);
        Mockito.when(authorizationService.isAnonAccessEnabled()).thenReturn(true);
        Mockito.when(authorizationService.isAnonymous()).thenReturn(true);
        Mockito.when(repositoryService.getAllRepoKeys()).thenReturn(Set.of("generic"));
        TestRestRequest restRequest = Mockito.mock(TestRestRequest.class);
        Mockito.when(restRequest.getServletRequest()).thenReturn(Mockito.mock(HttpServletRequest.class));
        Mockito.when(restRequest.getPathParamByKey(NativeBrowserResource.REPO_PATH)).thenReturn("generic/somePath");
        List<BaseBrowsableItem> expectedList = new LinkedList<>();
        Mockito.when(repositoryService.localFederatedCachedRepoDescriptorByKey("generic"))
                .thenReturn(new LocalRepoDescriptor());
        Answer<List<BaseBrowsableItem>> answer = invocationOnMock -> {
            RootNodesFilterResult rootNodesFilterResult = invocationOnMock.getArgument(2,
                    RootNodesFilterResult.class);
            rootNodesFilterResult.setAllItemNodesCanRead(true);
            return expectedList;
        };
        Mockito.doAnswer(answer).when(repoBrowsingService)
                .getLocalRepoBrowsableChildren(any(BrowsableItemCriteria.class), Mockito.anyBoolean(),
                        any(RootNodesFilterResult.class));

        HttpServletResponse httpServletResponse = new MockHttpServletResponse();
        ArtifactoryRestResponse<NativeBrowserModel> response = new ArtifactoryRestResponse<>(httpServletResponse);
        browseNativeService.execute(restRequest, response);

        Assertions.assertEquals(HttpStatus.SC_UNAUTHORIZED, response.getResponseCode());
    }

    @Test
    public void testWithUnauthorizedAnonUserWhenAnonAccessEnabled() {
        Mockito.when(authorizationService.canRead(any(RepoPath.class))).thenReturn(false);
        Mockito.when(authorizationService.isAnonAccessEnabled()).thenReturn(true);
        Mockito.when(authorizationService.isAnonymous()).thenReturn(true);
        Mockito.when(repositoryService.getAllRepoKeys()).thenReturn(Set.of("generic"));
        TestRestRequest restRequest = Mockito.mock(TestRestRequest.class);
        Mockito.when(restRequest.getPathParamByKey(NativeBrowserResource.REPO_PATH)).thenReturn("generic/somePath");
        Mockito.when(repositoryService.localFederatedCachedRepoDescriptorByKey("generic"))
                .thenReturn(new LocalRepoDescriptor());
        Mockito.when(restRequest.getServletRequest()).thenReturn(Mockito.mock(HttpServletRequest.class));

        List<BaseBrowsableItem> expectedList = new LinkedList<>();
        Answer<List<BaseBrowsableItem>> answer = invocationOnMock -> {
            RootNodesFilterResult rootNodesFilterResult = invocationOnMock.getArgument(2,
                    RootNodesFilterResult.class);
            rootNodesFilterResult.setAllItemNodesCanRead(false);
            return expectedList;
        };
        Mockito.doAnswer(answer).when(repoBrowsingService)
                .getLocalRepoBrowsableChildren(any(BrowsableItemCriteria.class), Mockito.anyBoolean(),
                        any(RootNodesFilterResult.class));

        HttpServletResponse httpServletResponse = new MockHttpServletResponse();
        ArtifactoryRestResponse<NativeBrowserModel> response = new ArtifactoryRestResponse<>(httpServletResponse);
        browseNativeService.execute(restRequest, response);

        Assertions.assertEquals(HttpStatus.SC_UNAUTHORIZED, response.getResponseCode());
    }

    @Test
    public void testWithUnExistRepoKeyAndAuthorizedUser() {
        Mockito.when(authorizationService.canRead(any(RepoPath.class))).thenReturn(true);
        Mockito.when(repositoryService.getAllRepoKeys()).thenReturn(Set.of(""));
        TestRestRequest restRequest = Mockito.mock(TestRestRequest.class);
        Mockito.when(restRequest.getPathParamByKey(NativeBrowserResource.REPO_PATH)).thenReturn("generic/somePath");
        ArtifactoryRestResponse<NativeBrowserModel> response = new ArtifactoryRestResponse<>();
        browseNativeService.execute(restRequest, response);

        Assertions.assertEquals(HttpStatus.SC_NOT_FOUND, response.getResponseCode());
    }

    @Test
    public void testWithUnExistRepoKeyAndAnonUser() {
        Mockito.when(authorizationService.canRead(any(RepoPath.class))).thenReturn(true);
        Mockito.when(authorizationService.isAnonAccessEnabled()).thenReturn(true);
        Mockito.when(authorizationService.isAnonymous()).thenReturn(true);
        Mockito.when(repositoryService.getAllRepoKeys()).thenReturn(Set.of(""));
        TestRestRequest restRequest = Mockito.mock(TestRestRequest.class);
        Mockito.when(restRequest.getPathParamByKey(NativeBrowserResource.REPO_PATH)).thenReturn("generic/somePath");
        ArtifactoryRestResponse<NativeBrowserModel> response = new ArtifactoryRestResponse<>();
        browseNativeService.execute(restRequest, response);

        Assertions.assertEquals(HttpStatus.SC_UNAUTHORIZED, response.getResponseCode());
    }

    @Test
    public void testWithFileDownload() {
        RepoPath repoPath = RepoPathFactory.create("generic/folder/file");
        Mockito.when(repositoryService.getAllRepoKeys()).thenReturn(Set.of(repoPath.getRepoKey()));
        Mockito.when(authorizationService.canRead(any(RepoPath.class))).thenReturn(true);
        TestRestRequest restRequest = Mockito.mock(TestRestRequest.class);
        Mockito.when(restRequest.getServletRequest()).thenReturn(Mockito.mock(HttpServletRequest.class));
        ArtifactoryRestResponse<NativeBrowserModel> response = new ArtifactoryRestResponse<>();
        Mockito.when(restRequest.getPathParamByKey(NativeBrowserResource.REPO_PATH)).thenReturn(repoPath.toPath());
        Mockito.when(repositoryService.localFederatedCachedRepoDescriptorByKey(repoPath.getRepoKey()))
                .thenReturn(new LocalRepoDescriptor());
        Mockito.when(repoBrowsingService
                .getLocalRepoBrowsableChildren(any(BrowsableItemCriteria.class), Mockito.anyBoolean(),
                        any(RootNodesFilterResult.class))).thenThrow(new FolderExpectedException(repoPath));

        browseNativeService.execute(restRequest, response);
        NativeBrowserModel model = response.getIModel();
        Assertions.assertEquals("generic", model.getRepo());
        Assertions.assertEquals("folder/file", model.getPath());
        Assertions.assertFalse(model.isRoot());
        Assertions.assertFalse(model.isFolder());
        Assertions.assertNull(model.getChildren());
        Assertions.assertNull(model.getFooter());
    }

    @Test
    public void testWithRealRepoPath() {
        Mockito.when(repositoryService.getAllRepoKeys()).thenReturn(Set.of("generic"));
        Mockito.when(authorizationService.canRead(any(RepoPath.class))).thenReturn(true);
        TestRestRequest restRequest = Mockito.mock(TestRestRequest.class);
        Mockito.when(restRequest.getServletRequest()).thenReturn(Mockito.mock(HttpServletRequest.class));
        ArtifactoryRestResponse<NativeBrowserModel> response = new ArtifactoryRestResponse<>();
        List<BaseBrowsableItem> expectedList = new LinkedList<>();
        expectedList
                .add(new BrowsableItem("folder", true, 123L, 1234L, 200L, RepoPathFactory.create("generic", "folder")));
        expectedList
                .add(new BrowsableItem("file", false, 124L, 1235L, 30L, RepoPathFactory.create("generic", "file")));
        Mockito.when(restRequest.getPathParamByKey(NativeBrowserResource.REPO_PATH)).thenReturn("generic/somePath");
        Mockito.when(repositoryService.localFederatedCachedRepoDescriptorByKey("generic"))
                .thenReturn(new LocalRepoDescriptor());
        Mockito.when(repoBrowsingService
                .getLocalRepoBrowsableChildren(any(BrowsableItemCriteria.class), Mockito.anyBoolean(),
                        any(RootNodesFilterResult.class))).thenReturn(expectedList);
        browseNativeService.execute(restRequest, response);

        Assertions.assertEquals(HttpStatus.SC_OK, response.getResponseCode());
        NativeBrowserModel model = response.getIModel();
        Assertions.assertEquals("generic", model.getRepo());
        Assertions.assertEquals("somePath", model.getPath());
        Assertions.assertEquals("1.0.0 Server at null Port 0", model.getFooter());
        Assertions.assertFalse(model.isRoot());
        Assertions.assertTrue(model.isFolder());
        List<NativeBrowserChild> items = model.getChildren();
        Assertions.assertEquals("folder", items.get(0).getName());
        Assertions.assertEquals(200L, items.get(0).getSize());
        Assertions.assertEquals("file", items.get(1).getName());
        Assertions.assertEquals(30L, items.get(1).getSize());
    }

    @Test
    public void testWithRealRepoPathAndAuthorizedAnonUser() {
        Mockito.when(authorizationService.canRead(any(RepoPath.class))).thenReturn(false);
        Mockito.when(authorizationService.isAnonAccessEnabled()).thenReturn(true);
        Mockito.when(authorizationService.isAnonymous()).thenReturn(true);
        Mockito.when(repositoryService.getAllRepoKeys()).thenReturn(Set.of("generic"));
        TestRestRequest restRequest = Mockito.mock(TestRestRequest.class);
        Mockito.when(restRequest.getPathParamByKey(NativeBrowserResource.REPO_PATH)).thenReturn("generic/somePath");
        Mockito.when(repositoryService.localFederatedCachedRepoDescriptorByKey("generic"))
                .thenReturn(new LocalRepoDescriptor());
        Mockito.when(restRequest.getServletRequest()).thenReturn(Mockito.mock(HttpServletRequest.class));

        List<BaseBrowsableItem> expectedList = new LinkedList<>();
        expectedList
                .add(new BrowsableItem("folder", true, 123L, 1234L, 200L, RepoPathFactory.create("generic", "folder")));
        Answer<List<BaseBrowsableItem>> answer = invocationOnMock -> {
            RootNodesFilterResult rootNodesFilterResult = invocationOnMock.getArgument(2,
                    RootNodesFilterResult.class);
            rootNodesFilterResult.setAllItemNodesCanRead(false);
            return expectedList;
        };
        Mockito.doAnswer(answer)
                .when(repoBrowsingService)
                .getLocalRepoBrowsableChildren(any(BrowsableItemCriteria.class), Mockito.anyBoolean(),
                        any(RootNodesFilterResult.class));

        HttpServletResponse httpServletResponse = new MockHttpServletResponse();
        ArtifactoryRestResponse<NativeBrowserModel> response = new ArtifactoryRestResponse<>(httpServletResponse);
        browseNativeService.execute(restRequest, response);

        Assertions.assertEquals(HttpStatus.SC_OK, response.getResponseCode());
        NativeBrowserModel model = response.getIModel();
        Assertions.assertEquals("generic", model.getRepo());
        Assertions.assertEquals("somePath", model.getPath());
        Assertions.assertEquals("1.0.0 Server at null Port 0", model.getFooter());
        Assertions.assertFalse(model.isRoot());
        Assertions.assertTrue(model.isFolder());
        List<NativeBrowserChild> items = model.getChildren();
        Assertions.assertEquals(1, items.size());
    }

    @Test
    public void testWithFileAndUnAuthorizedAnonUser() {
        Mockito.when(authorizationService.canRead(any(RepoPath.class))).thenReturn(false);
        Mockito.when(authorizationService.isAnonAccessEnabled()).thenReturn(true);
        Mockito.when(authorizationService.isAnonymous()).thenReturn(true);
        Mockito.when(repositoryService.getAllRepoKeys()).thenReturn(Set.of("virtual"));
        TestRestRequest restRequest = Mockito.mock(TestRestRequest.class);
        Mockito.when(restRequest.getPathParamByKey(NativeBrowserResource.REPO_PATH))
                .thenReturn("virtual/somePath/file.zip");
        Mockito.when(repositoryService.virtualRepoDescriptorByKey("virtual"))
                .thenReturn(new VirtualRepoDescriptor());
        Mockito.when(repoBrowsingService
                .getVirtualRepoBrowsableChildren(any(BrowsableItemCriteria.class), any(Boolean.class),
                        any(RootNodesFilterResult.class)))
                .thenThrow(new FolderExpectedException(RepoPathFactory.create("virtual", "somePath/file.zip")));
        Mockito.when(restRequest.getServletRequest()).thenReturn(Mockito.mock(HttpServletRequest.class));


        HttpServletResponse httpServletResponse = new MockHttpServletResponse();
        ArtifactoryRestResponse<NativeBrowserModel> response = new ArtifactoryRestResponse<>(httpServletResponse);
        browseNativeService.execute(restRequest, response);

        Assertions.assertEquals(HttpStatus.SC_UNAUTHORIZED, response.getResponseCode());
    }

    private static class TestRestRequest extends ArtifactoryRestRequest<String> {
        public TestRestRequest(RequestBuilder<String> requestBuilder) {
            super(requestBuilder);
        }
    }
}